"""

This package has the following modules:
    
``user_input_checking``
    This module contains functions for checking user input.

``tracker_schema``
    This module contains the schema for validating user input.
    
``athr_srch_modularized``
    This module contains functions to complete the author_search command modularized into pieces.
    
``athr_srch_webio``
    This module contains functions for author_search to interface with the internet.
    
``athr_srch_emails_and_reports``
    This module contains functions to create emails and reports for author_search.
    
``ref_srch_modularized``
    This module contains functions to complete the reference_search command modularized into pieces.
    
``ref_srch_webio``
    This module contains functions for reference_search to interface with the internet.
    
``ref_srch_emails_and_reports``
    This module contains functions to create emails and reports for reference_search.
    
``citation_parsing``
    This module contains functions for parsing references and citations for reference_search.

``fileio``
    This module contains functions for reading and writing files.
    
``helper_functions``
    This module contains functions that help the other modules function. The functions do things such as fuzzy matching, regex searching, and printing.
        
``webio``
    This module contains general functions for interfacing with the internet.

"""
__version__ = "1.0.2"
